from os.path import (
    exists,
    join,
    dirname,
    basename,
    splitext,
    expanduser,
    relpath,
    isdir
)
from .compiler import Ref, Schedule
from os import environ, walk
import configparser, json, os, yaml
import logging
from functools import lru_cache
from pathlib import Path
from .gits.module import GitModule
from .utils import system


class Settings:
    # dev utils
    debug = False

    # if path provided leads to a repo upstream
    is_repo      : bool = False
    is_module    : bool = False
    is_ymvas     : bool = False # if repos if fetched from ymvas
    is_main      : bool = False # if repo is main git@ymvas.com/vas/vas.git

    # paths
    root         : str  = None # /root             -> folder of the repo
    git          : str  = None # /root/.git        -> folder , if module -> /root/../../.git
    hooks        : str  = None # /root/.git/hooks  -> folder , if module -> /root/../../.git/hooks
    ymvas        : str  = None # /root/.ymvas      -> folder

    # repo info
    url          : str  = None # example  -> git@ymvas.com/vas/ymvas.git
    name         : str  = None # from url -> ymvas
    user         : str  = None # from url -> vas
    alias        : str  = None # from url -> vas/ymvas

    # ymvas important folders
    d_references : str = None
    d_commands   : str = None
    d_settings   : str = None
    d_tasks      : str = None
    d_schedules  : str = None
    d_secrets    : str = None
    d_finance    : str = None
    d_hooks      : str = None
    d_endpoints  : str = None
    d_documents  : str = None
    d_contacts   : str = None
    d_spaces     : str = None

    # where are we going to complie the endpointss
    d_compile_prefix : str = None

    # submodules asociated
    modules: dict = {}

    # ymvas important files
    f_settings            : str = None
    f_settings_endpoints  : str = None
    f_settings_references : str = None
    f_settings_secrets    : str = None
    f_settings_tasks      : str = None
    f_settings_finance    : str = None
    f_settings_hooks      : str = None

    d_default_compiles    : str = None


    # private
    __git_modules = []
    __references  = {}

    def __init__( self , pwd ):
        root, is_repo = Settings._find_repo_root(pwd)
        self.__logger = logging.getLogger(__name__)

        self.root = root
        if is_repo:
            self._setup_repo_paths()

        self.is_repo = is_repo

        data = system.get_global_config()
        if "version" not in data:
            os.makedirs(
                dirname( system.global_config_dir() ),
                exist_ok = True
            )

            with open(system.global_config_file(),'w') as fs:
                from .__init__ import __version__

                data = {
                    "version" : __version__ ,
                }

                if self.is_main:
                    data['global-src']      = self.root
                    data['global-commands'] = self.d_commands

                fs.write(json.dumps(data, indent= 10))
        else:
            data = system.get_global_config()
            self.d_default_compiles = data.get('global-compile-dir',None)

    def log(self,*args,color=None,**kwargs):

        if self.debug:
            if color != None:
                painter = {
                    "red"    : lambda x: f"\033[31m{x}\033[0m",
                    "green"  : lambda x: f"\033[32m{x}\033[0m",
                    "yellow" : lambda x: f"\033[33m{x}\033[0m",
                    "blue"   : lambda x: f"\033[34m{x}\033[0m",
                }.get(color, lambda x: x)
                args = [painter(a) for a in args]

            print(*args)
            return

        self.__logger.info(*args,**kwargs)

    @property
    @lru_cache(maxsize=None)
    def ymvas_server_url(self):
        stg = system.get_global_config()
        return str(stg.get(
            'ymvas-server-url',
            'ssh://git@'+self.ymvas_domain+'/{repo}.git'
        ))

    @property
    @lru_cache(maxsize=None)
    def is_server(self):
        return system.get_global_config().get( 'is-ymvas', False )

    @property
    @lru_cache(maxsize=None)
    def ymvas_domain( self ):
        stg = system.get_global_config()
        return str(stg.get(
            'ymvas-domain' ,
            'ymvas.com'
        ))

    @staticmethod
    def _find_repo_root(pwd):
        _pwd_b = pwd

        ymf = join(pwd , '.ymvas' )
        ymp = join(pwd , '.ymvas' )
        git = join(pwd , '.git'   )

        _is_repo = True
        _search  = {}

        while not exists(ymf):
            _pwd_b = dirname(_pwd_b)

            _search[_pwd_b] = _search.get(_pwd_b,0) + 1
            ymf = join(_pwd_b,'.ymvas')

            if _search[_pwd_b] > 10 or exists(git):
                ymf = ymp
                _is_repo = False
                break

            git = join( _pwd_b , '.git'   )
            ymp = join( _pwd_b , '.ymvas' )

        return dirname(ymf), _is_repo

    def _setup_repo_paths( self ):
        _git      = join( self.root , '.git' )
        is_module = not isdir( _git )

        if not exists(_git):
            return

        if is_module:
            with open( _git ,'r' ) as f:
                _git = Path(f.read().split("gitdir:")[1].strip())
                _git = str( Path(self.root) / _git )

        self.git       = _git
        self.hooks     = join(_git,'hooks')
        self.is_module = is_module

        ######### config #########
        cnf = configparser.ConfigParser()
        cnf.read(join( _git , 'config' ))

        _url = None
        for section in cnf.sections():
            if 'origin' in section and 'remote':
                _url = cnf[section].get('url',None)
                break


        user, name, is_ymvas = self._parse_url(_url)

        self.is_ymvas = is_ymvas
        self.user     = user
        self.name     = name
        self.url      = _url

        self.is_main = user == name
        self.alias   = f'{name}'
        if user is not None:
            self.alias = f'{user}/{name}'

        # /repo/.ymvas folders
        self.ymvas = join(self.root, '.ymvas' )

        self.d_references = join( self.ymvas, 'references' )
        self.d_commands   = join( self.ymvas, 'commands'   )
        self.d_settings   = join( self.ymvas, 'settings'   )
        self.d_tasks      = join( self.ymvas, 'tasks'      )
        self.d_schedules  = join( self.ymvas, 'schedules'  )
        self.d_secrets    = join( self.ymvas, 'secrets'    )
        self.d_finance    = join( self.ymvas, 'finance'    )
        self.d_hooks      = join( self.ymvas, 'hooks'      )
        self.d_endpoints  = join( self.ymvas, 'endpoints'  )

        # /repo folders if account
        if self.is_main:
            self.d_endpoints = join(self.root, 'endpoints' )
            self.d_documents = join(self.root, 'documents' )
            self.d_contacts  = join(self.root, 'contacts'  )
            self.d_finance   = join(self.root, 'finance'   )
            self.d_spaces    = join(self.root, 'spaces'    )

        # ymvas endpoints paths
        self.d_compile_prefix = self.alias
        if self.is_main:
            self.d_compile_prefix = user

        self.modules   = self.get_modules()

        # sttings files
        self.f_settings            = join(self.d_settings, 'settings.yaml'   )
        self.f_settings_endpoints  = join(self.d_settings, 'endpoints.yaml'  )
        self.f_settings_references = join(self.d_settings, 'references.yaml' )
        self.f_settings_secrets    = join(self.d_settings, 'secrets.yaml'    )
        self.f_settings_tasks      = join(self.d_settings, 'tasks.yaml'      )
        self.f_settings_finance    = join(self.d_settings, 'finance.yaml'    )
        self.f_settings_hooks      = join(self.d_settings, 'hooks.yaml'      )
        self.f_settings_schedules  = join(self.d_settings, 'schedules.yaml'  )

    @lru_cache(maxsize=None)
    def relpath(self,path):
        return relpath(path,self.root)

    def refs(self, src = None):
        _dir = self.d_references

        if src is not None:
            _dir = join(self.d_references,src)

        for r,_,_files in walk(_dir):
            for f in files:
                yield Ref(join(ff), self)

    def get_ref(self,space:str,fragment:str) -> Ref:
        _exists = self.__references.get(space,{}).get(fragment,None)

        if _exists:
            return _exists

        # get refs only for active or in scope of user
        module = self.get_modules().get(space,{})
        m_path = join(module.path , '.ymvas', 'references' )

        if not module.path or not module.active:
            return

        if not exists(m_path):
            return

        for r,_,files in walk( m_path ):
            for f in files:
                ff = join(r,f)
                if not Ref.match(ff,fragment):
                    continue

                r = Ref( ff , self )
                # store for later use
                if space not in self.__references:
                    self.__references[space] = {fragment:r}
                else:
                    self.__references[space][fragment] = r
                return r

        pass

    
    def schedules(self):
        for r,_,files in walk( self.d_schedules ):
            for f in files:
                schedule = Schedule(join(r,f))
                if not schedule.active:
                    continue
                yield schedule

        if self.is_main:
            for r,_,files in walk( self.d_contacts ):
                for f in files:
                    schedule = Schedule(join(r,f))
                    schedule.transform_contact()

                    if not schedule.active:
                        continue

                    yield schedule

    @lru_cache(maxsize=None)
    def get_modules(self):
        file = join(self.root,'.gitmodules')

        if self.is_module:
            modules = join(".git","modules")
            fragment = modules + self.git.split(modules)[-1]
            main_path = self.git.replace(fragment,"")
            main_settings = Settings(main_path)
            return main_settings.get_modules()

        m = GitModule()
        m.root = True
        m.active = True
        m.path = self.root
        m.url  = self.url
        m.name = self.name
        m.user = self.user

        modules = { self.alias : m }

        if not exists(file):
            return modules

        cnf = configparser.ConfigParser()
        cnf.read( file )

        for s in cnf.sections():
            p = cnf[s].get('path'   , None )
            u = cnf[s].get('url'    , None )
            a = cnf[s].get('active' , 'True' ).lower().strip() == 'true'

            if not 'submodule' in s or p is None or u is None:
                continue

            user, name, is_ymvas = self._parse_url(u)
            if name is None:
                continue

            m = GitModule()
            m.root = False
            m.active = a
            m.path = join( self.root , p )
            m.url = u
            m.name = name
            m.user = user

            modules[f"{user}/{name}"] = m

        return modules

    @lru_cache(maxsize=None)
    def get_commands(self, is_global = False, filter = None):
        dr = self.d_commands
        if is_global or not self.is_repo:
            dr = system.get_global_config().get( 'global-commands',None )

        if dr is None: return

        # widnows git bash fixes
        dr = Path(dr)
        if str(dr).startswith("/c/") or str(dr).startswith("\\c\\"):
            dr = Path("C:" + str(dr)[2:])

        if not dr.exists():
            return

        valid = [
            {"ext" : 'py'   , "run" : "python3" },
            {"ext" : 'bash' , "run" : "bash"    },
            {"ext" : 'sh'   , "run" : "sh"      }
        ]

        vdict = {x['ext']:x['run'] for x in valid}

        if filter is None:
            for r,_,files in os.walk(dr):
                if len(files) == 0:
                    continue
                for f in files:
                    ab = join(r,f)
                    fl = str(Path( ab ).relative_to( dr ))
                    st = splitext(fl)
                    rn = vdict.get(st[1].strip('.'),None)

                    if rn is None:
                        continue

                    yield {
                        "cmd"  : st[0],
                        "run"  : rn,
                        "path" : ab
                    }
            return

        for t in valid:
            file = join(dr,f"{filter}." + t['ext'])
            if exists(file):
                yield {
                    "cmd"  : filter,
                    "run"  : t['run'],
                    "path" : file
                }
                break

 
    @lru_cache(maxsize=None)
    def get_ymvas_settings(self):
        return self.__json_file(self.f_settings)

    @lru_cache(maxsize=None)
    def get_ymvas_hooks_settings(self):
        return self.__yaml_file(self.f_settings_hooks)

    @lru_cache(maxsize=None)
    def get_ymvas_schedules_settings(self):
        return self.__yaml_file(self.f_settings_schedules)

    def __json_file(self,file, types = dict):
        if not exists(file):
            return {}
        try:
            with open(file,'r') as f:
                data = json.loads(f.read())
                if not isinstance(data,types):
                    return {}
                return data
        except Exception:
            return {}
        return {}

    def __yaml_file(self,file, types = dict):
        if not exists(file):
            return {}
        try:
            with open(file,'r') as f:
                data = yaml.safe_load(f.read())
                if not isinstance(data,types):
                    return {}
                return data
        except Exception as e:
            return {}
        return {}

    def set_global_settings(self,stg):
        with open(system.global_config_file(),'w') as fs:
            fs.write(json.dumps(stg))

    def _parse_url(self, url ):
        if url is None:
            return None, None, None

        is_ymvas = f"@{self.ymvas_domain}" in url
        name = basename(url).replace('.git','')

        # get repo user
        user1 = url.split('@')[-1]
        user  = user1.replace(basename(url),'')
        user  = basename(user.strip('/'))
        user  = user.split(':')[-1]
        user  = user.strip('/')
        user  = user if '/' not in user else None

        if user is not None:
            user  = None if user1.startswith( user ) else user

        return user, name, is_ymvas




def mk_html_pdf(file, content, conf:dict = {}):

    options = {
      'margin-top'    : str(conf.get( 'margin-top'    , '0.0in' )),
      'margin-right'  : str(conf.get( 'margin-right'  , '0.0in' )),
      'margin-bottom' : str(conf.get( 'margin-bottom' , '0.0in' )),
      'margin-left'   : str(conf.get( 'margin-left'   , '0.0in' )),
    }

    try:
        import pdfkit
        
        file.write(pdfkit.from_string(
            content,
            False,
            options = options
        ))
        return True
    except OSError as e:
        file.write(content.encode('utf-8'))
        print(
           "[pdfkit] failed to create file, "
           "please install all the depedencies required!"
        )
        return False


def mk_svg_png(file,content,conf:dict = {}):
    _content = content.encode('utf-8')
    try:
        from cairosvg import svg2png
        file.write(svg2png(bytestring=_content))
        return True
    except Exception as e:
        file.write(_content)

        print(
            '[cairosvg] failed to create file, '
            'please install all the depedencies required!'
        )
        return False

def mk_md_html(file,content,conf:dict = {}):
    _content = content.encode('utf-8')
    try:
        import markdown
        file.write(markdown.markdown(content).encode('utf-8'))
        return True
    except Exception as e:
        file.write(_content)
        print(
            '[markdown] failed to create file, '
            'please install all the depedencies required!'
        )
        return False


def make(expect:str,into:str, file, content, conf:dict = {}):
    if   expect == "html" and into == 'pdf':
        return mk_html_pdf(file, content , conf )
    elif expect == 'svg'  and into == 'png':
        return mk_svg_png( file, content , conf )
    elif expect == 'md'   and into == 'html':
        return mk_md_html( file, content , conf )

    return False



import os, json, shutil, sys
from os.path import join, exists, dirname


class Gits:
    _separator_start = "# ymvas [start] don't change this line"
    _separator_end = "# ymvas [end] don't change this line"

    def __init__(self, stg):
        self.stg = stg

        # skip server side, non repo and non ymvas
        if stg.is_server or \
           not stg.is_repo or \
           not stg.is_ymvas:
            return

        self._conf_file = join(stg.hooks,'ymvas.json')
        self._check_hocks()

    def _check_hocks(self):
        stg = self.stg

        conf = self._get_ymvas_hooks()
        self._new_hock_conf()

        from ..__init__ import __version__
        if conf.get('version') != __version__:
            self._new_hock_conf()
            return


    # @lru_cache(maxsize=None)
    def _get_ymvas_hooks(self):
        try:
            with open(self._conf_file,'r') as f:
                data = json.loads(f.read())
                if not isinstance(data,dict):
                    return {}
                return data
        except Exception as e:
            return {}

    def _new_hock_conf(self):
        hook = join(self.stg.hooks,'pre-commit' )

        content = ""
        if exists(hook):
            with open(hook,'r') as f:
                content = f.read()

        skip = False
        ncontent = []
        for l in content.split("\n"):
            if self._separator_end in l:
                skip = False
                continue
            if skip:
                continue
            if self._separator_start in l:
                skip = True
                continue
            ncontent.append(l)

        root = "$(git rev-parse --show-toplevel)"
        ncontent = [
            "#!/bin/sh",
            self._separator_start,
            f"{sys.executable} -m ymvas trigger pre-commit --src=\"{root}\" ",
            self._separator_end,
        ] + ncontent[1:]

        with open(hook,'w') as f:
            f.write("\n".join(ncontent))

        # sys.executable
        with open(self._conf_file,'w') as fs:
            from ..__init__ import __version__

            data = {
                "version" : __version__ ,
            }

            fs.write(json.dumps(data, indent= 10))

    def pull(self):
        pass



import json, os, platform
from functools import lru_cache


@lru_cache(maxsize=None)
def sys_platform():
    system  = platform.system()
    shell   = os.environ.get("SHELL", "")
    msystem = os.environ.get("MSYSTEM", "")
    
    # Detect Git Bash on Windows
    is_windows_git_bash = (system == "Windows") and (
        "bash" in shell or msystem.startswith("MINGW")
    )

    if is_windows_git_bash:
        return "git-bash"
    elif system == "Windows":
        return "widnows"
    elif system == "Darwin":
        return "mac"
    return 'unix'


@lru_cache(maxsize=None)
def global_config_dir():
    return os.path.join({
        "git-bash" : os.path.expanduser("~/.config"),
        "widnows"  : os.environ.get("APPDATA", os.path.expanduser(
            "~\\AppData\\Roaming"
        )),
        "mac"      : os.path.expanduser("~/Library/Application Support"),
        'unix'     : os.environ.get("XDG_CONFIG_HOME", os.path.expanduser(
            "~/.config"
        ))
    }.get(sys_platform(), '~/.config'), 'ymvas')


@lru_cache(maxsize=None)
def global_config_file():
    return os.path.join(global_config_dir(),'config.json')


@lru_cache(maxsize=None)
def get_global_config():
    f = global_config_file()
    if not os.path.exists(f):
        return {}
    try:
        with open(f) as fs:
            return json.loads(fs.read())
    except Exception:
        return {}
    return {}





import yaml, uuid
from functools import lru_cache
from croniter import croniter
from ics import Event
from os.path import basename
from datetime import datetime as dt
from ics.grammar.parse import ContentLine
from datetime import date, datetime as dt
from dateutil import parser 


class Schedule:

    def __init__(self,path):
        self.path = path 
        self.data = {}
        self._is_contact = False
    
    def transform_contact(self):
        self._load()
        cont = self.data
        birthday = cont.get("birthday")

        self.data = {
            "active" : True,
            'is-all-day' : 'true'
        }

        if birthday is not None:
            bdate = Schedule._date(birthday)
            self.data['cron'] = f"0 0 {bdate.day} {bdate.month} *"
            self.data['start'] = birthday
        else:
            self.data['active'] = False 

        self._is_contact = True
    
    @property
    def is_contact(self):
        return self._is_contact
    
    def validate(self):
        ics = self.data.get("ics",{})
        if str(ics.get('active','false')).lower() == 'false':
            return
        
        if 'ics' in self.data and isinstance(self.data['ics'],dict):
            self.data['ics']['active'] = 'true'
        elif 'ics' in self.data and not isinstance(self.data['ics'],dict):
            self.data['ics'] = {'active':'true'}
        elif 'ics' not in self.data:
            self.data['ics'] = {'active':'true'}


    @lru_cache
    def _load(self):
        try:
            with open(self.path,'r') as f:
                self.data = yaml.safe_load(f.read())
        except Exception:
            pass

    @property
    def name(self):
        bn = basename(self.path)
        bn = bn.split('.')[:-1]
        bn = ".".join(bn)
        return bn

    @property
    @lru_cache
    def active(self):
        self._load()
        return str(self.data.get('active',False)).lower() == 'true'
    
    @property
    @lru_cache
    def is_ics(self):
        ics = self.data.get('ics',{})
        ics_active = str(ics.get('active',"False")).lower()
        ics_active = ics_active == 'true' or ics_active == '1'
        return isinstance(ics,dict) and ics_active

    @property
    def cron_expr(self):
        return self.data.get('cron',None)

    @property
    @lru_cache
    def cron(self):
        if self.cron_expr is None:
            return

        if not croniter.is_valid(self.cron_expr):
            return

        return croniter(self.cron_expr,dt.now())
        
    def ics_days(self,days):
        _days = ['SU','MO','TU','WE','TH','FR','SA']
        return ','.join([_days[(int(d))] for d in days])

    @property
    @lru_cache
    def valid(self):
        return self.cron is not None \
                and self.is_ics \
                and self.cron_data is not None
    
    @property
    @lru_cache
    def cron_data(self):
        if self.cron is None:
            return None
    
        return croniter.expand(self.cron_expr)
    
    @staticmethod
    def _date(dx,default=date(1990,1,1)):
        if isinstance(dx,str):
            try:
                px = parser.parse(dx).date()
                if px is None:
                    return default
                return px
            except Exception:
                return default
        elif isinstance(dx,dict):
            y,m,d = dx.get('year'), dx.get("month"),dx.get("day")

            if y == None or \
               m == None or \
               d == None or \
               not str(y).isnumeric() or \
               not str(m).isnumeric() or \
               not str(d).isnumeric():
                return default
                
            return date(int(y),int(m),int(d))
        return default
    

    @property
    @lru_cache
    def start_date(self):
        data = self.data
        _date = data.get("start",data.get("start-date"))
        return Schedule._date(_date,date(1900,1,1))


    @property
    @lru_cache
    def end_date(self):
        data = self.data
        _date = data.get("end" ,data.get("end-date"))
        return Schedule._date(_date,date(3000,1,1))
       

    @property
    @lru_cache
    def urgency(self):
        u = self.data.get('urgency',0)
        if str(u).isnumeric():
            return u
        return 0

    @property
    def is_all_day(self):
        return str(self.data.get('is-all-day','false')).lower() == 'true'

    @property
    def event(self):
        e = Event()

        e.name        = self.name
        e.uid         = str(uuid.uuid4())
        e.description = self.data.get('description','undefined')
        e.priority    = self.urgency
        e.begin       = self.start_date

        parts         = self.cron_data[0]

        if self.is_all_day:
            e.make_all_day()

        yearly = self.data.get("yearly",None)
        if yearly is not None and str(yearly).isnumeric():
            days = ",".join(str(x) for x in parts[2])
            mont = ",".join(str(x) for x in parts[3])

            e.extra.append(ContentLine(
                name  = 'RRULE',
                value = f"FREQ=YEARLY;INTERVAL={yearly};BYMONTH={month};BYMONTHDAY={days}"
            ))

        elif parts[2] != ["*"] and parts[4] == ["*"]:
            rule = "FREQ=MONTHLY;BYMONTHDAY=" + ",".join(str(x) for x in parts[2])
            e.extra.append(ContentLine(name="RRULE", value=rule))
        elif parts[4] != '*':
            rule = "FREQ=WEEKLY;BYDAY=" + self.ics_days(parts[4])
            e.extra.append(ContentLine(name="RRULE", value=rule))
        else:
            rule = "FREQ=DAILY"
            e.extra.append(ContentLine(name="RRULE", value=rule))

        return e



from functools import lru_cache


class GitModule:
    root   : bool = False
    active : bool = False
    path   : str  = None
    url    : str  = None
    name   : str  = None
    user   : str  = None

    @property
    def alias(self):
        return f"{self.user}/{self.name}"
    
    @property
    @lru_cache(maxsize=True)
    def settings(self):
        from .. import Settings 
        return Settings(self.path)




import requests

class API:
    url = 'http://api.ymvas.com'

    def __init__(
        self,
        auth:str ,
        is_ssh:bool = False
    ):

        self.auth = auth
        pass


    def post(self, fragment, *args,**kwargs):
        return request.post( self.url + fragment , *args, **kwargs )

    def tmp_add_key(self, name, content ):
        self.post( '/tmp_add_key' ,
            data = {
                "name" : name,
                "content" : content ,
                "pasw" : self.auth,
            }
        )



from functools import lru_cache
from ..utils import system


class Arguments:
    is_version  : bool = False
    is_global   : bool = False
    debug       : bool = False

    command     : str  = None
    # if command is not set in commands it means it's an executable
    exec        : str  = None

    action      : str  = None
    repo        : str
    init        : str

    src         : str
    compile_dir : str

    commands = [
        "info"    ,
        "clone"   ,
        "config"  ,
        "pull"    ,

        "trigger" , # triggers a hook
        "secret"  , # get secret 
        "compile" , # compile references and endpoints
        "init"    , # create the template for ymvas
    ]

    _config_args = [
        "global-compile-dir" , # where the compiles will be stored
        "global-commands"    , # where the global commands are stored
        "global-src"         , # where the user repo is stored

        # server settings
        'is-ymvas'           , # if it's server side or not [true or false]
        'ymvas-server-url'   , # url typically https://ymvas.com
        'ymvas-compile-url'  , # url typically https://docs.ymvas.com
        'ymvas-domain'       , # domain typically ymvas.com
        'ymvas-access'       , # access privileges
    ]


    def __init__( self , command:str,  *args , src = None ):
        _, self.is_version  = Arguments.flag(
            [command]+list(args),'-v','--version', False
        )

        setattr(
            self,
            'command' if command in self.commands else 'exec',
            command
        ) # first statement

        # flags [non required]
        args, self.debug       = Arguments.flag(args,'-d','--debug'  , False )
        args, self.is_global   = Arguments.flag(args,'-g','--global' , False )
        args, self.src         = Arguments.keyv(args,'-s','--src'    , src   )
        args, self.compile_dir = Arguments.keyv(args,'-c','--compile-dir'    )

        if self.src is None and self.is_global:
            cnf = system.get_global_config()
            self.src = cnf.get('global-src',None)

        self.args = args
        self._get_2nd_arg()

    def is_valid(self):
        pass

    def _get_2nd_arg(self):
        getters = {
            "clone"   : lambda a: a[0] if '/' in a[0] else f"{a[0]}/{a[0]}",
        }

        _2nd = getters.get( self.command, lambda a: a[0] )
        _2nd = None if len( self.args ) == 0 else _2nd(self.args)

        setters = {
            "clone"   : "repo",
            "init"    : "init",
            "config"  : "action",
            "trigger" : "action",
        }

        if self.command in setters and _2nd is not None:
            setattr( self, setters[self.command] , _2nd )
            self.args = self.args[1:]

    @staticmethod
    def flag(args,short_key,key,default=False):
        trash, value, found = [], default, False
        for i, c in enumerate( args ):
            if ( c == key or c == short_key ) and not found:
                value, found = True, True
            else:
                trash.append(c)
        return trash, value

    @staticmethod
    def keyv(args,short_key,key,default=None):
        trash, value, found = [], default, False
        for i, c in enumerate(args):
            if (c.startswith(key) or c.startswith(short_key)) and not found:
                found = True
                value = c.split('=')[0]
                value = c.replace(value + '=',"")
                if len(value) > 2 and value[0] == value[-1] and value[0] in ["'",'"']:
                    value = value.strip(value[0])
            else:
                trash.append(c)
        return trash, value

    @lru_cache(maxsize=None)
    def get_config_args(self):
        cnfs = {}
        for arg in self._config_args:
            self.args, value = Arguments.keyv(self.args,f'--{arg}',f'--{arg}')
            if value is None or value == '':
                continue
            cnfs[arg] = value
        return cnfs

    @lru_cache(maxsize=None)
    def get_config_flags(self):
        cnfs = []
        for arg in self._config_args:
            self.args, value = Arguments.flag(self.args,f'--{arg}',f'--{arg}')
            if value != True:
                continue
            cnfs.append(arg)
        return cnfs



from ymvas.compiler.references import Ref
from ymvas.settings import Settings
from os.path import exists, dirname, isdir, join, basename
import inspect, os, json, yaml, io
from dotenv.main import DotEnv


class Environ:

    def __init__(self, src = None, settings = None):
        self.src = src

        if src is None:
            caller   = inspect.stack()[1]
            self.src = dirname(caller.filename)

        self.settings = settings
        if settings is None:
            self.settings = self.__get_settings()
        
        self.__is_env = False
        self.file = self.__find__file()
        self.ref = None

        if not (self.file is None):
            self.ref = Ref(self.file,self.settings)

    def __get_settings(self):
        return Settings(self.src)

    def __find__file(self):

        if exists(self.src) and not isdir(self.src):
            dna = basename(self.src)
            self.__is_env = (dna == '.env')
            return self.src

        posible = [
            ['.env'     , True],
            [".env.json", False],
            [".env.yaml", False],
            [".env.yml" , False],
        ]

        for f,is_env in posible:
            ff = join(self.src,f)
            if not exists(ff) or isdir(ff):
                continue
            self.__is_env = is_env
            return ff

        for f,is_env in posible:
            ff = join(self.settings.root,f)
            if not exists(ff) or isdir(ff):
                continue
            self.__is_env = is_env
            return ff
            
    def __load_env(self):
        e = DotEnv(
            None,
            stream = io.StringIO(self.ref.content)
        )
        return e.dict()

    def __parse(self):
        if self.ref is None:
            return {}

        cnt = self.ref.content

        if self.__is_env:
            return self.__load_env()
        elif self.ref.lang == 'json':
            return json.loads(cnt)
        elif self.ref.lang == 'yaml':
            return yaml.safe_load(cnt)
        elif self.ref.lang == 'yml':
            return yaml.safe_load(cnt)
        
        return {}

    def load(self):
        data = self._parse()
        if not isinstance(data,dict):
            return
        for k, v in data.items():
            os.environ[k] = str(v)

    def __get__(self,key):
        return os.environ.get(key)

    def get(self,key,default=None):
        return os.environ.get(key,default)



from functools   import lru_cache
from os.path import join, relpath, isdir, basename, dirname
from os import remove, makedirs, sep

import shutil
from .modifiers  import make
from .references import Ref


class Change:
    
    DELETED = 'D'
    UPDATED = 'U'

    def __init__(self,settings, file,action):
        self.origin = file
        self.action = action
        self.settings = settings
        
        # populated on sync old
        self._prev_compiled = None
        self._destination = None

        self.format = None
        self.into = None
    
    @property
    @lru_cache(maxsize=None)
    def fullpath(self):
        return join(self.settings.root, self.origin)

    @property
    @lru_cache(maxsize=None)
    def is_endpoint(self):
        endpoints_path = self.settings.relpath( self.settings.d_endpoints )
        return self.origin.startswith( endpoints_path )
    
    @property
    @lru_cache(maxsize=None)
    def is_setting(self):
        settings_path = self.settings.relpath( self.settings.d_settings )
        return self.origin.startswith( settings_path )
    
    @property
    @lru_cache(maxsize=None)
    def is_endpoints_setting(self):
        file = self.settings.relpath(self.settings.f_settings_endpoints)
        return self.origin == file
    
    
    @property
    @lru_cache(maxsize=None)
    def cnf_name(self):
        if self.is_setting:
            return relpath(self.fullpath,self.settings.d_settings)
        return relpath(self.fullpath, self.settings.d_endpoints )
 
    @property
    @lru_cache(maxsize=None)
    def cnf_end(self):
        name = self.cnf_name
        if not self.into:
            return name
        
        base = basename(name)
        frgm = name.strip(base)
        fnam = [base] if not '.' in base else base.split('.')[:-1]
        fnam = ".".join(fnam)
        fnam+= "." + self.into

        return frgm + fnam

    @property
    def end_basename(self):
        return basename(self.cnf_end)
    
    @property
    def cnf_dir(self):
        dd = dirname(self.cnf_name)
        if dd is None:
            return self._destination
        return join(self._destination,dd)

    @property
    def is_dir(self):
        return isdir(self.fullpath)
    
    def sync(self,destination, conf):
        self._destination = destination

        if not self.is_endpoint:
            return
        
        files    = conf.get('files',{})
        current  = files.get(self.cnf_name,{})
        compiled = current.get('compiled',None)

        if compiled is not None:
            self._prev_compiled = join(destination,compiled)
        
        cnf_e = current.get('conf',{})
        self.format = cnf_e.get('format',None)
        self.into   = cnf_e.get('into',None)

    def rm(self):
        if self._prev_compiled:
            remove(self._prev_compiled)
            self.settings.log(f"Deleted {self._prev_compiled}")
        
    def is_outdated(self):
        return self.action == Change.DELETED
    
                
    def create(self):
        e = Ref( self.fullpath , self.settings )
        makedirs( self.cnf_dir , exist_ok = True )
        
        destination = join(self.cnf_dir, self.end_basename )
        
        if e.just_copy:
           shutil.copy2( self.fullpath , destination )
           self.settings.log(f"file {self.origin} was copied")
           return
            
        with open(destination, 'wb') as f:
            cnf = {}
            if self.format and self.into:
                cnf['format'] = self.format
                cnf['into']   = self.into
            
            success = make(
               self.format,
               self.into,
               f,
               e.content,
            )

            if not success:
                f.write(e.content.encode('utf-8'))

    
    def config(self):
        cnf = {}
        if self.format and self.into:
            cnf['format'] = self.format
            cnf['into'] = self.into


        return (
            self.cnf_name,
            {
                "origin" : self.origin,
                "compiled" : self.cnf_end,
                "conf" : cnf
            }
        )



import os

from os.path     import join, exists
from .compiler   import Ref, Compiler
from .settings   import Settings
from .templates  import Creator
from .gits       import Gits
from .utils      import system 

class Operator:
    _refs  :dict = {}
    _spaces:dict = {}

    def __init__(self, pwd , debug = False):
        self.settings       = Settings( pwd )
        self.settings.debug = True
        # self.git            = Gits(self.settings)

    def command(self,args):
        cmd = list(self.settings.get_commands(
            args.is_global,
            args.exec
        ))

        if len(cmd) == 0:
            print(f'Command [{args.exec}] is not setup for this repository!')
            print(self)
            return

        cmd = cmd[0]
        os.system(cmd['run'] + " " + cmd['path'])


    def clone(self,args):
        if args.repo is None:
            print('No repo specified!')
            return

        url = self.settings.ymvas_server_url.format(
            repo = args.repo
        )

        os.system(f"git clone {url}")

    def pull(self,argv):
        modules = self.settings.get_modules()
        modules = {k:v for k,v in modules.items() if not v.root and v.active}

        os.system(
            f"git --git-dir={self.settings.git} "
            f"--work-tree={self.settings.root} fetch"
        )

        # for k,v in modules.items():
        #     p = v[ 'path' ]
        #     u = v[ 'url'  ]
        #
        #     if not exists(p):
        #         os.system(
        #             f"cd {self.settings.root} && "
        #             f"git --git-dir={self.settings.git} "
        #             f"--work-tree={self.settings.root} "
        #             f"submodule add {u} {p}"
        #         )


    def push(self):
        if not self.settings.is_server:
            print('is not server', self.settings.hooks )


    def setup(self,argv):
        creator = Creator(argv)
        creator.run()

    def config(self,argv):
        if argv.is_global:
            stg = system.get_global_config()
            if argv.action == 'set':
                stg = stg | argv.get_config_args()
                self.settings.set_global_settings(stg)
            elif argv.action == "show":
                print(stg)
            elif argv.action == "get":
                for a in argv.get_config_flags():
                    v = stg.get(a,None)
                    print(v)
            return

    def secret(self,argv):
        pass

    def trigger(self,argv):
        hooks = self.settings.get_ymvas_hooks_settings()
        hook  = hooks.get( argv.action , None )
        if hook is None:
            return
        exec = hook.get("exec",None)
        file = hook.get('src',None)
        if not exec or not file:
            return
        file = join(self.settings.root,file)
        if not exist(file):
            return
        os.system(f'{exec} {file}')


    def compile(self,args):
        Compiler( 
            self.settings ,
            join( args.compile_dir, self.settings.d_compile_prefix )
        ).run()


    def __repr__( self ):
        data = [
            f"[{self.settings.alias}]",
            f" - is-repo  : {self.settings.is_repo}",
            f" - is-ymvas : {self.settings.is_ymvas}",
            f" - is-main  : {self.settings.is_main}",
            "",
            f"[global configuration]",
            f" - {system.global_config_file()}"
        ]

        cmds = [c['cmd'] for c in self.settings.get_commands()]
        if len(cmds) != 0:
            data += [
                "", f"[avaliable commands]",
                " - " +  "\n - ".join(cmds),
            ]

        return "\n".join(data) + "\n"










    #



import os, re
from os.path import join, basename, exists
from .soup import Soup


class Ref:
    cli       = None
    raw:str   = ''
    soup:Soup = None
    _content  = None
    _bloked   = False

    def __init__(self, fpath , stg ):
        self.space    = stg.alias
        self.settings = stg
        self.fpath    = fpath
        self.basename = basename(fpath)

        parts         = self.basename.split(".")
        self.name     = parts[0]
        self.lang     = 'xml' if len(parts) == 1 else parts[-1]
        
        fallbacks = {
            'yml' : "yaml"
        }

        self.lang = fallbacks.get(self.lang,self.lang)

        # for non readable files
        self.just_copy = False
        try:
            with open(self.fpath, 'rb') as f:
                while chunk := f.read(4096):
                    chunk.decode('utf-8')
        except UnicodeDecodeError:
            self.just_copy = True
    
    @staticmethod
    def match( file:str , link:str ):
        bn = basename(file)
        np = bn.split('.')
        nm = np[0]
        if link == nm:
            return True
        en = "" if len(np) == 0 else np[-1]
        nm = file.strip(en)
        return '/' in link and nm.endswith(link)

    def __repr__( self ):
        return f"<ref link='{self.name}' />"

    def _compile(self):
        if self._content != None:
            return

        with open(self.fpath,'r') as f:
            self.raw = f.read()

        self.soup    = Soup( self.raw , logger = self.settings.log )
        self._bloked = True

        for t in self.soup.refs():
            self.settings.log(f"     {t}" , color = 'green' )
            space = t.get('space', self.settings.alias )
            name  = t.ref()

            ref   = self.settings.get_ref( space ,name )

            if ref is None:
                self.settings.log(f"     {t} not found" , color = 'red' )
                continue

            if ref is self:
                t.set_replace(ref.raw,ref.lang)
                continue

            t.set_replace(ref.content,ref.lang)

         
        self._content = self.soup.to_string()
        self._bloked  = False

    @property
    def content(self):
        if self._bloked:
            soup = Soup(self.raw)
            for t in soup.refs():
                t.set_replace('')
            return soup.to_string()
        self._compile()
        return self._content



import sys, os
from os.path import join, exists, dirname
from .operator import Operator
from ymvas.models.arguments import Arguments


_defined_config   = ['set','get','show']
_defined_inits    = ['account','default']

red    = lambda x: f"\033[31m{x}\033[0m"
green  = lambda x: f"\033[32m{x}\033[0m"
yellow = lambda x: f"\033[33m{x}\033[0m"
blue   = lambda x: f"\033[34m{x}\033[0m"


def _default_info():
    print(f"{green('YMVAS')} - handle your life like a developer!")
    print(f"Usage : "                                 )

    print(f"    {green('ym')} {{command}}   |  {yellow(', '.join(Arguments.commands))} ")
    print(f"    {green('ym')} --version   |  {green('Is allowed!')}              ")
    print( ""                                         )

    print( "File based commands:"                     )
    print( "    {repo}/.ymvas/commands/install.bash"  )
    print( "    {repo}/.ymvas/commands/install.py "   )
    print( "    {repo}/.ymvas/commands/install.sh "   )
    print( ""                                         )

    print(f"Can be easly executed like this: "        )
    print(f"    {green('ym')} install        "        )
    print( ""                                         )
    print(f"Flags:                           "        )
    print(f"    [ {yellow('--debug')}  ] or [{yellow('-d')}]" )
    print(f"    [ {yellow('--global')} ] or [{yellow('-g')}]" )

def _compile_info():
    print(f"Usage : ")
    print(f"    {green('ym')} compile -c=\"~/destination/folder\"")
    print(f"    {green('ym')} compile --compile-dir=\"~/destination/folder\"")
    print(f"")
    print(yellow(f"You can also setup default compile directory with: "))
    print(yellow(f"    ym config set --global-compile-dir=\"~/your/path\""))

def _compile_no_repo():
    print(red(f"[{os.getcwd()}]: is not a valid repository!"))
    print("")

def _config_info():
    print(f"Usage : ")
    print(f"    {green('ym')} config     | {yellow(', '.join(_defined_config))} ")
    print(f"    {green('ym')} config -g  | For global configuration ")
    print(f"")
    print(f"Example : ")
    print(f"    {green('ym')} config set -g  | For global configuration ")

def _init_info():
    print(f"Usage : ")
    print(f"    {green('ym')} init     | {yellow(', '.join(_defined_inits))} ")
    print(f"")
    print(f"Example : ")
    print(f"    {green('ym')} init account ")



def run():
    args = sys.argv[1:]

    # no arguments passed
    if len(args) == 0:
        _default_info()
        sys.exit(1)

    argv = Arguments(args[0], *args[1:], src = os.getcwd())

    if argv.is_version: # print version
        from .__init__ import __version__
        print( __version__ )
        sys.exit(0)

    # extract src
    cli = Operator( argv.src , debug = argv.debug )

    if argv.command == 'info':
        print( cli )
        return

    if argv.command == "clone":
        cli.clone( argv )
        return

    if argv.command == "init":
        if argv.init is None:
            _init_info()
            return
        cli.setup( argv )
        return

    if argv.command == 'config':
        if argv.action not in _defined_config:
            _config_info()
            return
        cli.config(argv)
        return

    if argv.exec is not None:
        cli.command(argv)
        return


    # found repo
    if not cli.settings.is_repo:
        _compile_no_repo()
        print(cli)
        return

    if argv.command == 'compile':
        if cli.settings.d_default_compiles is None and argv.compile_dir is None:
            _compile_info()
            return

        cli.compile( argv )
        return

    if argv.command == 'trigger':
        cli.trigger(argv)
        return

    if argv.command == 'secret':
        cli.secret(argv)

    if argv.command == 'pull':
        cli.pull(argv)
        return


if __name__ == '__main__':
    run()



from os.path import join, exists, dirname, basename, isdir, relpath
from pathlib import Path
import os, json, shutil, yaml, subprocess
from os import walk, sep, remove

from functools import lru_cache
from ..settings import Settings
from .change import Change
from ics import Calendar

class Compiler:

    def __init__( self, settings:Settings , destination:str ):
        self.settings    = settings
        self.destination = destination
        
        self.modules = [
                m.alias for m in settings.get_modules().values() if not m.root
        ]

        self.config = {}
        
        self.file = join(destination,'config.json')
        if exists(self.file):
            with open(self.file) as f:
                self.config = json.loads(f.read())
        
        self.modules_changed = False

    @property
    @lru_cache(maxsize=None)
    def _git(self):
        return (
            f"git --git-dir='{self.settings.git}' "
            f" --work-tree='{self.settings.root}' "
        )
    
    @property
    @lru_cache(maxsize=None)
    def commit(self):
        return os.popen(
            f"{self._git} rev-parse HEAD"
        ).read().strip()
    
    @property
    @lru_cache(maxsize=None)
    def first_commit(self):
        return os.popen(
            f"{self._git} rev-list --max-parents=0 HEAD"  
        ).read().strip()
    
    @property
    @lru_cache(maxsize=None)
    def is_uncommited(self):
        return os.popen(f"{self._git} status -s").read().strip() != ""
    

    def __compiled_commit(self):
        comm = self.config.get('commit',self.first_commit)
        from ..__init__ import __version__

        if self.config.get('version','0') != __version__ \
           or self.is_uncommited:
            return self.first_commit
        
        return comm
        
    @property
    @lru_cache(maxsize=None)
    def compiled_commit(self):
        cm = self.__compiled_commit()
        return cm if self.is_forced else cm
        
    @property
    @lru_cache(maxsize=None)
    def is_forced(self):
        compiled_commit = self.__compiled_commit()
        if compiled_commit == self.first_commit:
            return True
        try:
            subprocess.run([
                "git", 
                f"--git-dir='{self.settings.git}'",
                'merge-base', 
                '--is-ancestor', 
                compiled_commit, 
                self.commit
               ],
               check=True,
               capture_output = True,
               text=True 
            )
            return False
        except subprocess.CalledProcessError as e:
            return True
    


    def __changes(self, track = [], include_directories = False):
        if len(track) == 0:
            return []

        tracked = " ".join(track) # create the traking files
        
        # command changes commited
        cmd = (
           f"{self._git} --no-pager diff --name-status "
           f"{self.compiled_commit} {self.commit} -- {tracked}"
        )
        com = os.popen( cmd ).read().strip()
        
        cmd = f"{self._git} --no-pager status -s {tracked}"
        mod = os.popen( cmd ).read().strip()
        
        # get all changes in one list
        data    = mod.split("\n") + com.split("\n")
        changes = []

        for f in data:
            f = f.strip().replace("\t"," ")
            file  = " ".join(f.split(' ')[1:])

            if '"' in f or "'" in f:
                continue # skip strange files
            
            # handle rename 
            if f.upper().startswith("R"):
                fxf  = file.split('->')
                file = fxf[-1].strip()

                changes.append(Change(
                    self.settings, fxf[-1].strip(), Change.DELETED 
                ))
                changes.append(Change(
                    self.settings, fxf[0].strip() , Change.UPDATED
                ))

                continue

            if f.upper().startswith("D"):
                changes.append(Change( 
                    self.settings, file , Change.DELETED 
                ))
                continue

            changes.append(Change( self.settings, file , Change.UPDATED ))
        
        
        nchanges = []
        for c in changes:
            if c.is_dir and not include_directories: 
                continue
            c.sync(self.destination,self.config)
            nchanges.append(c)

        return nchanges
    
            
    def get_end_conf(self):
        if not exists(self.settings.f_settings_endpoints):
            return {}
        with open(self.settings.f_settings_endpoints,'r') as f:
            data = yaml.safe_load(f.read())
            if isinstance(data,dict):
                return data
        return {}


  
    def get_settings_changes(self,conf_e,conf_o):
        files = conf_o.get("files",{})
        endpoints_p = self.settings.relpath(
            self.settings.d_endpoints
        )
        
        changes = []
        for e, cnf_e in conf_e.items():
            old = files.get(e,{})

            if json.dumps(cnf_e) == json.dumps(old):
                continue

            file = join( endpoints_p, e )
            changes.append(Change(self.settings, file ,Change.UPDATED))
         
        return changes


    def clean(self):
        self.config = {}

        if not exists(self.destination):
            return True

        if not self.settings.is_main:
            self.settings.log(f"deleted all previous data")
            shutil.rmtree(self.destination)
            return True

        def _loop():
            for r,_,files in walk( self.destination ):
                for f in files:
                    yield join(r,f)

        for f in _loop():
            rf = relpath(f, self.destination )
            dd = rf.split( sep )
            if dd[0] in self.modules:
                continue
            self.settings.log(f"removed {rf}")
            remove(f)


        return True
    

    def run( self ):
        from ..__init__ import __version__

        if self.config.get('version','0') != __version__:
            self.settings.log('mismatch of version... removing all files')
            self.clean()
        
        elif self.is_uncommited:
            self.settings.log(f"Looks like working on local... removing all files")
            self.clean()

        elif self.is_forced:
            self.settings.log("Looks like forced push... removing all files")
            self.clean()

        self.settings.log(
            f"Different commit? {self.commit != self.compiled_commit}"
            f" | {self.commit} == {self.compiled_commit}"
        )

        if self.commit == self.compiled_commit:
            self.settings.log(f"No changes detected!")
            return
        
        
        config = {
            "commit"  : self.commit,
            "version" : __version__ 
        }
        
        os.makedirs( self.destination , exist_ok = True )

        # store modules folders so we don't overwrite them
        if self.settings.is_main:
            config['modules'] = self.modules
            for k,m in self.settings.get_modules().items():
                if len(self.__changes([m.path], include_directories=True)) != 0:
                    self.modules_changed = True
        
        # end preparation 
        self.__compile_settings(  config )
        self.__compile_schedules( config )
        self.__compile_endpoints( config )
        # self.__compile_tasks( config ) 

        
        with open(join(self.file),'w') as f:
            f.write(json.dumps(config, indent = 4))

        # print complie data
        self.settings.log(f"Final config : {json.dumps(config,indent = 1)}")
        return config

    

    def __compile_settings(self, config ):
        if not self.settings.is_main:
            return

        avatars = self.__changes([
            join(self.settings.d_settings,'avatar.jpeg'),
            join(self.settings.d_settings,'avatar.png'),
            join(self.settings.d_settings,'avatar.jpg'),
        ])
        

        ovatar = self.config.get('avatar',None)

        if len(avatars) != 0:
            avatar = avatars[0]
            avatar.sync( self.destination , self.config )
        
            ovapat = join(self.destination,str(ovatar))
            if ovatar != None and exists(ovapath):
                remove(ovapat)

            avatar.create()
            ovatar = avatar.cnf_name

        config['avatar'] = ovatar


    def __compile_schedules( self , config ):
        stg = self.settings
        rcf = stg.get_ymvas_schedules_settings()
        
        cal_path = join( self.destination , 'calendar.ics' )
        
        schedules = []
        include = rcf.get('include')
        contacts = rcf.get('contacts','none')
        
        if include is not None and self.modules_changed and stg.is_main:
            modules = []
            if isinstance(include,list):
                modules = include
            elif isinstance(include,dict):
                modules = include.get("modules",[])
                if not isinstance(modules,list):
                    modules = []

                modules = [m for m in modules if isinstance(m,str)]                
            elif isinstance(include,str):
                if include == '*' or include == 'all':
                    modules = self.modules
            
            gmodules = stg.get_modules()
            for m in modules:
                if m not in self.modules: 
                    continue
                
                md = gmodules[m]
                if not md.active: 
                    continue
                                    
                schedules += list(md.settings.schedules())


        # schedules changes
        changes = self.__changes([ stg.d_schedules ])
        
        cal = self.config.get('calendar',None)
        if len(changes) == 0 and not self.modules_changed:
            self.settings.log(
                "no changes detected in "
                "modules or schedules directory"
            )

            if cal != None:
                config['calendar'] = cal
            return
        
        schedules += list(self.settings.schedules())
        if len(schedules) == 0:
            self.settings.log("no schedules detected")
            return

        calendar = Calendar()
        for s in schedules:
            
            if s.is_contact and contacts == 'all':
                self.settings.log(f"{s.name} contact was added as birhtday")
                s.validate()

            if not s.valid: 
                self.settings.log(f"schedule [{s.name}] is not valid!")
                continue

            self.settings.log(f"schedule [{s.name}] will be added!")
            calendar.events.add(s.event)

        with open(cal_path,'w') as f:
            config['calendar'] = 'calendar.ics'
            self.settings.log("calendar.ics generated")
            f.writelines(calendar)


    def __compile_endpoints( self , config ):        
        changes = self.__changes([
          self.settings.d_endpoints ,
          self.settings.f_settings_endpoints
        ])

        if len(changes) == 0:
            self.settings.log(f"No endpoints changes detected!")
            config['files'] = self.config.get('files',[])
            return
            
        # sync and rm outdated, and get endpoints settings
        nchanges = []
        end_conf = {}
        avatar = None
        
        for c in changes:
            self.settings.log(c.fullpath)

            if c.is_endpoints_setting:
                end_conf = self.get_end_conf()
            
            if c.is_outdated() and c.is_endpoint:
                c.rm()
                continue
            elif not c.is_endpoint:
                continue
            nchanges.append(c)
        
        # check if settings are modified if so recompile the files changed
        schanges = self.get_settings_changes(end_conf,self.config)
        for c in schanges:
            c.sync(
                self.destination,
                self.config
            )

            _cnf = end_conf.get(c.cnf_name,{})

            self.settings.log(
                f"settings changes = {c.cnf_name} => "
                f"{_cnf}"
            )

            c.format = _cnf.get("format" ,None)
            c.into   = _cnf.get("into"   ,None)

            _exists = False
            for i, gc in enumerate(nchanges):
                if c.cnf_name == gc.cnf_name:
                    _exists = True
                    nchanges[i] = c
                    break

            if _exists:
                continue

            nchanges.append(c)
        
        config['files'] = {}
        for c in nchanges:
            c.create()
            name, data = c.config()
            config['files'][name] = data
        
